@@ -19,6 +19,8 @@ import sys
 import x_wr_timezone
 from collections import defaultdict
 from icalendar.prop import vDDDTypes
+from typing import Optional
+import re
 
 if sys.version_info[0] == 2:
     # Python2 has no ZoneInfo. We can assume that pytz is used.
@@ -33,12 +35,17 @@ if sys.version_info[0] == 2:
 else:
     def timestamp(dt):
         """Return the time stamp of a datetime"""
-        return dt.timestamp() 
+        return dt.timestamp()
+
+
+NEGATIVE_RRULE_COUNT_REGEX = re.compile(r"COUNT=-\d+;?")
+
 
 def convert_to_date(date):
     """converts a date or datetime to a date"""
     return datetime.date(date.year, date.month, date.day)
 
+
 def convert_to_datetime(date, tzinfo):
     """converts a date to a datetime"""
     if isinstance(date, datetime.datetime):
@@ -54,6 +61,7 @@ def convert_to_datetime(date, tzinfo):
         return datetime.datetime(date.year, date.month, date.day, tzinfo=tzinfo)
     return date
 
+
 def time_span_contains_event(span_start, span_stop, event_start, event_stop, comparable=False):
     """Return whether an event should is included within a time span.
 
@@ -156,7 +164,7 @@ class Repetition:
 
 class RepeatedComponent:
     """Base class for RepeatedEvent, RepeatedJournal and RepeatedTodo"""
-    
+
     def __init__(self, component, keep_recurrence_attributes=False):
         """Create an component which may have repetitions in it.
 
@@ -169,6 +177,7 @@ class RepeatedComponent:
         self.keep_recurrence_attributes = keep_recurrence_attributes
         self.exdates = []
         self.exdates_utc = set()
+        self.replace_ends = {}  # DTSTART -> DTEND # for periods
         exdates = component.get("EXDATE", [])
         for exdates in ((exdates,) if not isinstance(exdates, list) else exdates):
             for exdate in exdates.dts:
@@ -178,7 +187,13 @@ class RepeatedComponent:
         rdates = component.get("RDATE", [])
         for rdates in ((rdates,) if not isinstance(rdates, list) else rdates):
             for rdate in rdates.dts:
-                self.rdates.append(rdate.dt)
+                if isinstance(rdate.dt, tuple):
+                    # we have a period as rdate
+                    self.rdates.append(rdate.dt[0])
+                    self.replace_ends[timestamp(rdate.dt[0])] = rdate.dt[1]
+                else:
+                    # we have a date/datetime
+                    self.rdates.append(rdate.dt)
 
         self.make_all_dates_comparable()
 
@@ -202,7 +217,8 @@ class RepeatedComponent:
             rule.exdate(exdate)
         for rdate in self.rdates:
             rule.rdate(rdate)
-        rule.rdate(self.start)
+        if not self.rrule or not self.rrule.until or not compare_greater(self.start, self.rrule.until):
+            rule.rdate(self.start)
 
     def create_rule_with_start(self, rule_string):
         """Helper to create an rrule from a rule_string starting at the start of the component.
@@ -237,19 +253,21 @@ class RepeatedComponent:
 
     def rrulestr(self, rule_string):
         """Return an rrulestr with a start. This might fail."""
-        rule = rrulestr(rule_string, dtstart=self.start)
+        rule_string = NEGATIVE_RRULE_COUNT_REGEX.sub("", rule_string) # Issue 128
+        rule = rrulestr(rule_string, dtstart=self.start, cache=True)
         rule.string = rule_string
+        rule.until = until = self._get_rrule_until(rule)
+        if is_pytz(self.start.tzinfo) and rule.until:
+            # when starting in a time zone that is one hour off to the end,
+            # we might miss the last occurrence
+            # see issue 107 and test/test_issue_107_omitting_last_event.py
+            rule = rule.replace(until=rule.until + datetime.timedelta(hours=1))
+            rule.until = until
         return rule
 
-    _until = UNTIL_NOT_SET = "NOT_SET"
-    def get_rrule_until(self):
-        """Return the UNTIL datetime of the rrule or None is absent."""
-        if self._until is not self.UNTIL_NOT_SET:
-            return self._until
-        self._until = None
-        if self.rrule is None:
-            return None
-        rule_list = self.rrule.string.split(";UNTIL=")
+    def _get_rrule_until(self, rrule):
+        """Return the UNTIL datetime of the rrule or None if absent."""
+        rule_list = rrule.string.split(";UNTIL=")
         if len(rule_list) == 1:
             return None
         assert len(rule_list) == 2, "There should be only one UNTIL."
@@ -257,15 +275,15 @@ class RepeatedComponent:
         if date_end_index == -1:
             date_end_index = len(rule_list[1])
         until_string = rule_list[1][:date_end_index]
-        self._until = vDDDTypes.from_ical(until_string)
-        return self._until
+        until = vDDDTypes.from_ical(until_string)
+        return until
 
     def make_all_dates_comparable(self):
         """Make sure we can use all dates with eachother.
 
         Dates may be mixed and we have many of them.
         - date
-        - datetime without timezome
+        - datetime without timezone
         - datetime with timezone
         These three are not comparable but can be converted.
         """
@@ -306,12 +324,11 @@ class RepeatedComponent:
                 start = start.tzinfo.localize(start.replace(tzinfo=None))
                 # We could now well be out of bounce of the end of the UNTIL
                 # value. This is tested by test/test_issue_20_exdate_ignored.py.
-                until = self.get_rrule_until()
-                if until is not None and start > until and start not in self.rdates:
+                if self.rrule is not None and self.rrule.until is not None and start > self.rrule.until and start not in self.rdates:
                     continue
-            if self._unify_exdate(start) in self.exdates_utc:
+            if self._unify_exdate(start) in self.exdates_utc or start.date() in self.exdates_utc:
                 continue
-            stop = start + self.duration
+            stop = self.replace_ends.get(timestamp(start), start + self.duration)
             yield Repetition(
                 self.component,
                 self.convert_to_original_type(start),
@@ -337,7 +354,7 @@ class RepeatedComponent:
         """Whether this is a recurrence/modification of an event."""
         return "RECURRENCE-ID" in self.component
 
-    def as_single_event(self):
+    def as_single_event(self) -> Optional[Repetition]:
         """Return as a singe event with no recurrences."""
         for repetition in self.within_days(self.start, self.start):
             return repetition
@@ -351,7 +368,7 @@ class RepeatedEvent(RepeatedComponent):
         """Return DTSTART"""
         ## Arguably, it may be considered a feature that this breaks if no DTSTART is set
         return self.component['DTSTART'].dt
-    
+
     def _get_component_end(self):
         """Yield DTEND or calculate the end of the event based on DTSTART and DURATION."""
         ## an even may have DTEND or DURATION, but not both
@@ -377,7 +394,7 @@ class RepeatedTodo(RepeatedComponent):
         due = self.component.get('DUE')
         if due is not None:
             return due.dt
-        
+
         ## Assume infinite time span if neither is given
         ## (see the comments under _get_event_end)
         return datetime.date(*DATE_MIN)
@@ -388,9 +405,9 @@ class RepeatedTodo(RepeatedComponent):
         end = self.component.get('DUE')
         if end is not None:
             return end.dt
-        
+
         dtstart = self.component.get("DTSTART")
-        
+
         ## DURATION can be specified instead of DUE.
         duration = self.component.get("DURATION")
         ## It is no requirement that DTSTART is set.
@@ -398,13 +415,13 @@ class RepeatedTodo(RepeatedComponent):
         ## way to set DUE.
         if duration is not None and dtstart is not None:
             return self.component["DTSTART"].dt + duration.dt
-        
+
         ## According to the RFC, a VEVENT without an end/duration
         ## is to be considered to have zero duration.  Assuming the
         ## same applies to VTODO.
         if dtstart:
             return dtstart.dt
-        
+
         ## The RFC says this about VTODO:
         ## > A "VTODO" calendar component without the "DTSTART" and "DUE" (or
         ## > "DURATION") properties specifies a to-do that will be associated
@@ -430,7 +447,7 @@ class RepeatedJournal(RepeatedComponent):
     ## consider it to have zero duration if a timestamp is given.
     _get_component_end = _get_component_start
 
-    
+
 # The minimum value accepted as date (pytz + zoneinfo)
 DATE_MIN = (1970, 1, 1)
 # The maximum value accepted as date (pytz + zoneinfo)
@@ -442,7 +459,7 @@ class UnsupportedComponent(ValueError):
 
 class UnfoldableCalendar:
     '''A calendar that can unfold its events at a certain time.'''
-    
+
     recurrence_calculators = {
         "VEVENT": RepeatedEvent,
         "VTODO": RepeatedTodo,
@@ -563,6 +580,8 @@ class UnfoldableCalendar:
         for event_repetitions in self.repetitions:
             if event_repetitions.is_recurrence():
                 repetition = event_repetitions.as_single_event()
+                if repetition is None:
+                    continue
                 vevent = repetition.as_vevent()
                 add_event(vevent)
                 if not repetition.is_in_span(span_start, span_stop):
@@ -584,11 +603,10 @@ class UnfoldableCalendar:
 
 def of(a_calendar, keep_recurrence_attributes=False, components=["VEVENT"]):
     """Unfold recurring events of a_calendar
-    
+
     - a_calendar is an icalendar VCALENDAR component or something like that.
     - keep_recurrence_attributes - whether to keep attributes that are only used to calculate the recurrence.
     - components is a list of component type names of which the recurrences should be returned.
     """
     a_calendar = x_wr_timezone.to_standard(a_calendar)
     return UnfoldableCalendar(a_calendar, keep_recurrence_attributes, components)
-
